import importlib.util
import sys
import types
import typing as T
from dataclasses import dataclass
from pathlib import Path

from typing_extensions import Never


@dataclass
class MutableBool:
    value: bool

    def set(self, value: bool) -> None:
        self.value = value


def import_from_path(
    module_name: str,
    path: Path,
    preprocess: T.Optional[T.Callable[[bytes], bytes]] = None,
    trim_traceback: bool = False,
) -> types.ModuleType:
    spec = importlib.util.spec_from_file_location(module_name, path)
    if not (spec and spec.loader):
        raise TypeError(
            f"File loader for {path} was not found. That most likely means that the file extension is not valid."
        )
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module

    try:
        # A cute little trick to preprocess code before running it
        if preprocess is not None:
            original_source_to_code = spec.loader.source_to_code  # type: ignore

            def source_to_code(data: bytes, path, *, _optimize=-1):
                return original_source_to_code(preprocess(data), path, _optimize=_optimize)

            spec.loader.source_to_code = source_to_code  # type: ignore
            original_dont_write_bytecode = sys.dont_write_bytecode
            sys.dont_write_bytecode = True
            spec.loader.exec_module(module)
            sys.dont_write_bytecode = original_dont_write_bytecode
            spec.loader.source_to_code = original_source_to_code  # type: ignore
        else:
            spec.loader.exec_module(module)
    except SystemExit as e:
        raise
    except BaseException:
        if not trim_traceback:
            raise
        import traceback

        exc = traceback.format_exc()
        if "_call_with_frames_removed\n" in exc:
            exc = (
                "Traceback (most recent call last):\n"
                + exc[exc.find("_call_with_frames_removed\n") + len("_call_with_frames_removed\n") :]
            )
        sys.stderr.write(exc)
        raise

    return module


def __raise_exception__(exc: BaseException) -> Never:
    raise exc



import argparse
import importlib
import importlib.abc
import importlib.util
import os
import re
import subprocess
import sys
import typing as T
from pathlib import Path as P

import typer
from ideas import import_hook, main_hack

from .token_transformers import transform_source
from .util import MutableBool, __raise_exception__

__version__ = "1.1.3"
__all__ = ["P", "sh", "re", "typer", "sys", "os", "__pysh_check_returncodes__"]

PIPE = subprocess.PIPE
DEVNULL = subprocess.DEVNULL
__last_bash_cmd_returncode__ = 0
__pysh_check_returncodes__ = MutableBool(True)


def sh(
    cmd: T.Union[str, T.List[str]],
    extra_env: T.Dict[str, str] = {},
    pipe_stdout=False,
    **kwargs,
) -> subprocess.CompletedProcess:
    global __last_bash_cmd_returncode__
    kwargs["env"] = dict(**(kwargs.get("env") or os.environ), **extra_env)
    if pipe_stdout:
        kwargs["stdout"] = PIPE
        kwargs["stderr"] = subprocess.STDOUT
    result = subprocess.run(cmd, shell=True, text=True, check=__pysh_check_returncodes__.value, **kwargs)
    __last_bash_cmd_returncode__ = result.returncode
    return result


def add_hook():
    """Creates and automatically adds the import hook in sys.meta_path"""
    hook = import_hook.create_hook(
        hook_name=__name__,
        transform_source=transform_source,
        extensions=[".pysh"],
    )
    return hook


def parse_argv(argv: T.List[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(prog="pysh")
    parser.add_argument("script", help="The script to run", nargs="?", default=None)
    parser.add_argument("-c", "--compile", help="Compile the script without running", action="store_true")
    parser.add_argument("-o", "--output", help="The compilation output file/directory", default=None)
    return parser.parse_args(argv)


def main(argv: T.Optional[T.List[str]] = None) -> None:
    if argv is None:
        argv = sys.argv[1:]
    args = parse_argv(argv)
    import builtins

    globals_ = globals()
    for attr in __all__:
        setattr(builtins, attr, globals_[attr])
    builtins.__raise_exception__ = __raise_exception__  # type: ignore
    builtins.__pysh_check_returncodes__ = __pysh_check_returncodes__  # type: ignore

    if args.script is None:
        from ideas import console

        console.configure(transform_source=transform_source)
        console.start(prompt=">>> ", banner=f"Pysh Console [Python version: {sys.version}]")
        exit(0)
    elif not P(args.script).is_file():
        print("Expecting a path to the script", file=sys.stderr)
        exit(1)
    elif args.compile:
        new_source = transform_source(P(args.script).read_text())
        new_source = "from pysh import *\n" + new_source
        if args.output is None:
            file = P(args.script).with_suffix(".py")
        else:
            output = P(args.output)
            if output.is_dir():
                file = output / P(args.script).with_suffix(".py")
            else:
                file = output
        file.write_text(new_source)
    else:
        add_hook()
        if argv is sys.argv:
            argv[0] = argv.pop(1)

        module_name = args.script[: -len(".pysh")]
        main_hack.main_name = module_name
        module_path = P(args.script).resolve()
        sys.path.insert(0, str(module_path.parent))
        try:
            importlib.import_module(module_name)
        except Exception:
            import traceback

            exc = traceback.format_exc()
            seeked_str = f'  File "{module_path}"'
            if seeked_str in exc:

                exc = "Traceback (most recent call last):\n" + exc[exc.index(f'  File "{module_path}"') :]
                sys.stderr.write(exc)
                sys.exit(1)
            else:
                raise


if __name__ == "__main__":
    main()



import re
import tokenize
from collections import namedtuple
from io import StringIO
from typing import Any, Generator, List, Optional, Tuple, Union

try:
    from rich import print as pprint
except ImportError:
    pass


def pprint(*a, **kw):
    pass


"""
char *special_builtins[] =
{
  ":", ".", "source", "break", "continue", "eval", "exec", "exit",
  "export", "readonly", "return", "set", "shift", "times", "trap", "unset",
  (char *)NULL
};

/* The builtin commands that take assignment statements as arguments. */
char *assignment_builtins[] =
{
  "alias", "declare", "export", "local", "readonly", "typeset",
  (char *)NULL
};

char *localvar_builtins[] =
{
  "declare", "local", "typeset", (char *)NULL
};

/* The builtin commands that are special to the POSIX search order. */
char *posix_builtins[] =
{
  "alias", "bg", "cd", "command", "false", "fc", "fg", "getopts", "jobs",
  "kill", "newgrp", "pwd", "read", "true", "umask", "unalias", "wait",
  (char *)NULL
};
"""


SIMPLE_MAGICAL_COMMANDS = {"cd": 'os.chdir(f"{argv}")'}
COMPLEX_MAGICAL_COMMANDS = {"set", "exit"}
RE_SHELLVAR = re.compile(r"(?<!\\)\$[\w@#]+")


def convert_to_shell_var(varname: str) -> str:
    if varname.isdigit():
        return f"{{sys.argv[{varname}] if len(sys.argv) > int({varname}) else __raise_exception__(IndexError('Command line argument ${varname} was not passed'))}}"
    elif varname == "#":
        return "{len(sys.argv) - 1}"
    elif varname == "@":
        return "{' '.join(sys.argv[1:])}"
    elif varname == "?":
        return "{__last_bash_cmd_returncode__}"
    elif varname == "$":
        return "{os.getpid()}"
    else:
        return f"{{os.environ['{varname}']}}"


def convert_to_magical_command(cmd: str, argv: str) -> Optional[str]:
    if cmd in SIMPLE_MAGICAL_COMMANDS:
        return SIMPLE_MAGICAL_COMMANDS[cmd].format(cmd=cmd, argv=argv)
    elif cmd in COMPLEX_MAGICAL_COMMANDS:
        if cmd == "set" and argv.startswith(("+e", "-e")):
            return f'__pysh_check_returncodes__.set({argv.startswith("-e")})'
        elif cmd == "exit":
            if not argv:
                argv = "0"
            if not argv.isdigit():
                raise SyntaxError(
                    f"only integer arguments for the exit command are supported. Problematic command: {cmd} {argv}"
                )
            return f"sys.exit({int(argv)})"


def modify_tokens(tokens: List[tokenize.TokenInfo]) -> Generator[Tuple[int, str], None, None]:
    bash_line_started = False
    piped_bash_line_started = False
    cmd = None
    for i, token in enumerate(tokens):
        if bash_line_started:
            if token.type == tokenize.NEWLINE:
                if cmd is None:
                    raise SyntaxError(f"Unknown syntax error on line: {token.line}")
                prev_token = tokens[i - 1]
                sep = "!!" if piped_bash_line_started else "!"
                _, _, argv = prev_token.line.partition(sep + cmd)
                argv = argv.replace('"', '\\"').strip()
                for var in set(RE_SHELLVAR.findall(argv)):
                    argv = argv.replace(var, convert_to_shell_var(var[1:]))
                magical_command = convert_to_magical_command(cmd, argv.strip())
                if magical_command is not None:
                    yield tokenize.NAME, magical_command
                else:
                    command = f'''sh(f"""{cmd} {argv}""", pipe_stdout={piped_bash_line_started}).stdout'''
                    yield tokenize.NAME, command
                bash_line_started = piped_bash_line_started = False
                cmd = None
                yield token[:2]
            elif token.string == "!":
                if piped_bash_line_started:
                    raise SyntaxError(f'Three "!" in one line are not supported. Problematic line: {token.line}')
                else:
                    pprint("PIPE", token)
                    piped_bash_line_started = True
            elif cmd is None:
                pprint("CMD", token)
                cmd = token.string
            else:
                pprint("ADD", token)
        elif token.string == "!":
            pprint("START", token)
            bash_line_started = True
        else:
            pprint("STANDARD", token)
            yield token[:2]


def transform_source(source: str, **kwargs: Any) -> str:
    tokens = list(tokenize.generate_tokens(StringIO(source).readline))
    tokens = list(modify_tokens(tokens))
    for tok in tokens:
        pprint(tok)
    source = tokenize.untokenize(tokens)
    pprint(source)
    return source



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pysh']

package_data = \
{'': ['*']}

install_requires = \
['ideas>=0.0.37,<0.0.38',
 'tomlkit>=0.10.0,<0.11.0',
 'typer[all]>=0.4.0,<0.5.0',
 'typing-extensions>=4.2.0,<5.0.0']

entry_points = \
{'console_scripts': ['pysh = pysh.pysh:main']}

setup_kwargs = {
    'name': 'pysh',
    'version': '1.1.4',
    'description': 'A bunch of shortcuts and import magic to simplify scripting in python',
    'long_description': '# Pysh\nBash is said to be the opposite of riding the bike -- you have to re-learn it every time. So if we\'re doing anything more complex than running a few commands, chances are google will be needed. But if we switch to python, then the reverse is true -- python is a bit too cumbersome for simple scripts. I try to alleviate this problem by making it easier to use bash from python.\n\nIt was inspired by Jupyter\'s way of handling bash.\n\n## Installation\n```bash\npip install pysh\n```\n## Quickstart\n* If you have a hello_world.pysh file with the contents below, you can run it using `pysh hello_world.pysh`:\n```bash\n!echo Hello World\nprint("Hello World")\n```\n* Most bash is going to be one-to-one translatable to pysh by putting exclamation marks at the beginning of each line:\n```bash\n!ls\n!cp some_file some/place\n```\n* It has most of the variables (\\$@, \\$#, $1, $2, etc) you would expect:\n```bash\n!echo $*\n```\n* You can also access python variables from bash using f-string notation\n```bash\nmy_file = "~/some/file" + ".txt"\n!cp {my_file} {my_file + ".bak"}\n```\n* If you wish to get the output of any command into a python variable, you can use double exclamation mark:\n```bash\n# Note that the default output of ls is silenced when !! are used\nlines = !!ls -lA\nfor line in lines.splitlines():\n    print(line)\n```\n## Builtins\nI consider the most comfortable way to write python scripts is to import os, pathlib, sys, and re first. Thus, all of these modules are pre-imported. However, instead of "pathlib.Path", we have "P". \n\nI also import typer as I consider it to be the most concise way to write any command line application.\n\nIf you wish to customize how you call bash subprocesses, you can use the builtin "sh" function. It supports all the same arguments as subprocess.run but has text=True and shell=True by default.\n\n## Magical Commands\nWe use subprocess.run to run bash commands but that has a few quirks. If you modify the process info in any way within the subprocess (cd, set, unset, etc), this information will not change for the currently running process. Hence we have a few magical commands that do not create a subprocess but instead use roughly equivalent python constructs.\n\nHere\'s a list of all the magical commands and the approximate conversions we apply to them:\n| Original        | Converted                                 |\n| -----------     | -----------                               |\n| !cd arg         | os.chdir(arg)                             |\n| !set -e         | \\_\\_pysh_check_returncodes\\_\\_.set(True)  |\n| !set +e         | \\_\\_pysh_check_returncodes\\_\\_.set(False) |\n| !exit arg       | sys.exit(int(arg))                        |\n## Notes on syntax\n* Each bash call is parsed until the end of the line so they are closer to statements than expressions. Hence the following is not possible:\n```bash\nprint(!!ls)\n```\n* Because we use f-strings to interface between python and bash, you will have to escape non-formatting "{" and "}" by typing them twice. You will also have to wrap these characters in quotes because prior to converting bash into executable format, we tokenize the entire file as if it was valid python. I.e. Unclosed braces, parenthesis, etc will cause a tokenize.TokenError.\n```bash\n# The invalid ways to do it\n!echo {\n!echo {{\n\n# The valid way to do it\n!echo "{{"\n```\n## FAQ\n* If you have any questions, encounter any bugs or want to suggest/contribute new features, please, use the issues in the github repository',
    'author': 'Stanislav Zmiev',
    'author_email': 'szmiev2000@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Ovsyanka83/pysh',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
